[BITS 16]	; use 16-bit code
[ORG 7C00h]	; code is loaded by BIOS at address 7C00h
start:
jmp begin	; jump over the BIOS Parameter Block
;nop		; some BIOSes need this

BS_OEMName		DB "ROOOS1.0"	;
BPB_BytsPerSec	DW 512		; Bytes per sector
BPB_SecPerClus	DB 1			; Sectors per Cluster
BPB_RsvdSecCnt	DW 1 			; Number of Reserved sectors
BPB_NumFATs		DB 2			; Number of FAT copies
BPB_RootEntCnt	DW 224		; No. of root directory entries
BPB_TotSec16	DW 2880		; Total number of sectors
BPB_MediaId		DB 0F0h		; Disk media ID
BPB_FATSz16		DW 9			; No. of sectors in a FAT
BPB_SecPerTrk	DW 18			; Sectors per Track
BPB_NumHeads	DW 02			; Number of heads
BPB_HiddSec		DD 0			; Number of hidden sectors
BPB_TotSec32	DD 0			; Total no. of sectors if > 65536
; The next parts are not part of the official specification

BS_DrvNum		DB 0			; Drive Number (0 for floppy)
BS_Reserved1	DB 0			;
BS_BootSig		DB 29h 		;
BS_VolID		DD 0526F4F73h	; Volume Serial no. (RoOs in hex)
BS_VolLab		DB "ROOOS DISK "	; Volume Label
BS_FilSysType	DB "FAT12 "	; Type of filesystem

begin:
mov [BS_DrvNum],dl		; get drive number from dl (set by BIOS)

xor ax,ax		; set segment registers
mov es,ax		; to zero
mov ds,ax

; Reset FDD
int 13h		; ax = 0 means reset FDD

in al,92h
or al,02h
out 92h,al
;||
;| Load FAT at 0000:7E00h |
;|________________________|
mov si,FloppyErr	; Error message to  be printed if failed
mov bx,7E00h	; offset
mov dl,[BS_DrvNum]; Drive number
xor dh,dh		; Head 0
mov cl,2		; Sector 2
mov ch,0		; Track 0
mov ah,2		; Function number 2 (read sectors)
mov al,9		; read 9 sectors
int 13h		; call BIOS interrupt
jc @PErr		; Failed if Carry Flag set

;||
;| Load root directory at 0000:9000h |
;|___________________________________|
mov dl,[BS_DrvNum]
mov dh,1		; Head 1
mov cl,2		; Sector 2
mov ch,0		; Track 0
mov ah,2		; Read function
mov al,14		; Read 14 sectors
mov bx,9000h	; offset
int 13h
jc @PErr
jmp @continue

@PErr:
jmp PrintErr
;||
;| Find kernel from directory |
;|____________________________|
@continue:
mov di,9000h		; make edi point to the root directory

@findkrnl:
mov si,KernelFile		; point esi to filename
mov dx,di			; save di
mov cx,11			; length is 11 chars
repz cmpsb			; compare
mov di,dx			; get saved di value
jz @krnlfound		; jump if kernel entry was found
add di,32			; or get next root directory entry

dec word[BPB_RootEntCnt]; decrement variable
jnz @findkrnl		; exit if no more entries

mov si,KernelErr
jmp PrintErr
mov ax,0E0Dh
int 10h

@krnlfound:
mov ax,[di+01Ah]		; get the start of cluster from root dir entry
mov [ReadCluster],ax	; save it

;mov ax,[di+01Ch]	; get size of file in bytes from root dir entry
;shr ax,9		; divide by 512 (to get # of sectors)
;inc ax		; increment it to have atleast 1 sector
;mov cx,ax		; save it

mov si,FloppyErr	; set error message
mov ax,0FFFFh
mov es,ax
mov bx,10h

push bx		; save bx (will be used in algo)
push es
@read:
	pop es
	pop bx			; get bx
	mov ax,[ReadCluster]	; get cluster number
	add ax,31			; add 31 to get sector number
	call ReadSector		; read in the sector
	add bx,200h			; increment edi by 512 bytes to read next sector
	push bx			; save bx as it is used in the following code
	push es

	xor ax,ax
	mov es,ax
	mov ax,[ReadCluster]
	mov di,ax			;
	shr di,1			; di * (3/2)
	add di,ax			; 
	add di,7E00h		; add the address of FAT
	mov bx,[es:di]		; then load next cluster number from FAT
	and word[ReadCluster],1	; see if the old cluster was a full integer
	jnz @oddCluster		; else go to odd cluster
@evenCluster:
	and bx,0FFFh		; if whole integer AND with 0FFFh
	mov [ReadCluster],bx	; save it
	cmp bx,0FFFh		; check for End Of File
	jnz @read			; if not go back to read next sector
	jmp @endread		; or jump to end

@oddCluster:			; if not whole integer
	shr bx,4			; shift right 4 bits
	mov [ReadCluster],bx	; save it
	cmp bx,0FFFh		; check for End Of File
	jnz @read			; if not go to read next sector

@endread:
	mov dx,03F2h		; FDC Port
	xor al,al			; zero al i.e motors off and no IRQ/DMA
	out dx,al			; Turn off Floppy drive motor

	mov si,KrnlFound
	call PrintStr


;	call Enable_A20
;	jmp 0FFFFh:10h
;	jmp short $
	jmp startp
PrintErr:
	call PrintStr
	mov ax,0E0Dh
	int 10h

	xor ax,ax
	int 16h
	jmp 0FFFFh:00
	ret

;Location in es:bx and sector in ReadCluster
ReadSector:
	push es
	push bx
	xor dx,dx
	mov bx,[BPB_SecPerTrk]
	div bx
	inc dx	; DX is now the sector #
	mov cx,dx	; get sector # into CL
	xor dx,dx
	mov bx,[BPB_NumHeads]
	div bx	; AX = cylinder # -- DX = head #
	shl ah,6	; shift cylinder high bits
	mov ch,al	; move cylinder low bits into CH
	or cl,ah	; mask cylinder high bits into CL
	mov dh,[BS_DrvNum]
	xchg dh,dl	; move head number into DH

	pop bx
	pop es

	mov al,1
	mov ah,2

	int 13h
	jc PrintErr
	ret

PrintStr:
	mov ah,0Eh
	@strloop: 
		lodsb 
		test al,al 
		jz @endloop
		int 10h
		jmp @strloop 
	@endloop
		ret

startp:
cli
xor ax,ax
mov ds,ax
lgdt [gdt_desc]
mov eax,cr0
or eax,1
mov cr0,eax
jmp 08h:pmode

[BITS 32]
pmode:
	mov ax,10h	; Mul by 8 number of segment after null seg
	mov ds,ax
	mov es,ax

	mov ax,18h	; Stack Segment
	mov ss,ax
	mov esp,090000h	; set stack at a good place

;	mov byte[ds:0B8000h],'K'
;	mov byte[ds:0B8001h],0Dh

	jmp 08h:100000h

FloppyErr	db	"Bad Floppy",0
KernelErr	db	"Kernel error",0
KrnlFound	db	"Kernel Found",0

KernelFile	db	"OSKERNELCOM"
ReadCluster	dw	0

gdt:
gdt_null:
;	dd 0
;	dd 0
	dq 0.0	; Null Descriptor
gdt_code:
	dw 0FFFFh	; limit
	dw 0		; base
	db 0		; base mid
	db 10011010b; Flags		Bits Reverse
	db 11001111b; Attributes	    ,,
	db 0		; Base low
gdt_data:
	dw 0FFFFh
	dw 0
	db 0
	db 10010010b
	db 11001111b
	db 0
gdt_stack:
	dw 0FFFFh
	dw 0
	db 0
	db 10010010b
	db 11001111b
	db 0
;gdt_video:
;	dw 0FFFFh
;	dw 0
;	db 0
;	db 10010010b
;	db 11001111b
;	db 0
gdt_end:

gdt_desc:	dw gdt_end-gdt-1
		dd gdt

times 510-($-$$) db 0

db 055h,0AAh
