; Started work on this first on 20th January
; The started rebuilding on 22/01/2005 afternoon
; Completed a minimal FDC driver on 22/01/2005 evening
; Rebuilt AGAIN on 25/01/2005
[BITS 32]
FDC_DMA_Buffer	equ	7000h

FDC_DOR	equ	03F2h	;Digital Output Register(write-only
FDC_MSR	equ	03F4h	;Main Status Register(read-only)
FDC_DRS	equ	03F4h	;DataRate Select Register(read-only)
FDC_DATA	equ	03F5h	;Data Register(read-only)
FDC_DIR	equ	03F7h	;Digital Input Register(read-only)
FDC_CCR	equ	03F7h	;Configuration Control Register(write-only)

CMD_SPECIFY	equ	03h	;Specify Drive Timings
CMD_WRITE	equ	0C5h	;Write Data(Multi Track,MFM)
CMD_READ	equ	0E6h	;Read Data(MFM Mode(Double density),Multi track mode,Skip Deleted)
;CMD_READ	equ	66h	;Read Data(MFM, Skip Deleted)
CMD_CALIBRATE equ	07h	;Recalibrate drive
CMD_SENSEINT equ	08h	;Sense Interrupt status
CMD_FORMAT	equ	4Dh	;Format Track(+ MFM)
CMD_SEEK	equ	0Fh	;Seek Track
CMD_VERSION	equ	10h	;FDC Version

DISK_TRACKS		equ	80
DISK_HEADS		equ	2
DISK_SPT		equ	18
DISK_GAP3RW		equ	1Bh
DISK_GAP3FMT	equ	54h


FDC_IRQ_Handler:
	cli
	pushad

	mov dword[FDC_Done],1

	mov al,20h	; Send EOI to PIC
	out 20h,al	;

	popad
	iret

FDC_MotorOff:
	mov edx,FDC_DOR
	xor eax,eax		; turn off all motors and disable DMA/IRQ
	out dx,al
	mov dword[FDC_Motor_Flag],0
	ret

FDC_MotorOn:
	mov eax,[FDC_Motor_Flag]
	or eax,eax
	jne @fdcmon_exit

	mov edx,FDC_DOR
	mov eax,01Ch
	out dx,al
	mov dword[FDC_Motor_Timer],9	; 500 ms
@fdcmon_1:
	cmp dword[FDC_Motor_Timer],0
	jne @fdcmon_1

	mov dword[FDC_Motor_Flag],1
@fdcmon_exit:
	ret
FDC_Init:
	push ebp
	mov ebp,esp

	mov esi,FloppyMsg
	xor eax,eax
	int 30h

	call FDC_GetVersion
	call FDC_Reset

	mov esi,FloppyOk
	xor eax,eax
	int 30h

	leave
	ret

FDC_GetVersion:
	push ebp
	mov ebp,esp

	mov eax,CMD_VERSION
	call FDC_SendCmd
	call FDC_GetByte

	cmp eax,80h
	jne @fdcgv_1
	mov esi,NEC765Msg
	jmp @fdcgv_2
@fdcgv_1:
	mov esi,FDCMsg
@fdcgv_2:
	xor eax,eax
	int 30h

	leave
	ret

FDC_Reset:
	xor eax,eax
	mov [FDC_Done],eax

	mov edx,FDC_DOR	; Disable IRQ/DMA and turn off all motors
	out dx,al

	mov edx,FDC_CCR	; Configuration Control Register
	out dx,al

	mov eax,0Ch		; Enable IRQ/DMA
	mov edx,FDC_DOR
	out dx,al		; Output to DOR

	call FDC_WaitDone
	mov ecx,4		; Intel says we should send Sense Int 4 times
@fdcr_loop:			; after resetting
	push ecx
	call FDC_SenseInt
	pop ecx
	dec ecx
	jne @fdcr_loop

	mov ebx,1		; Seek to track 1
	call FDC_Seek	; 
	call FDC_Recalibrate
	mov dword[newdisk],0
	ret

FDC_Recalibrate:
	call FDC_MotorOn
	xor eax,eax
	mov [FDC_Done],eax

	mov ebx,CMD_CALIBRATE
	call FDC_SendCmd
	xor ebx,ebx
	call FDC_SendCmd

	call FDC_WaitDone
	call FDC_SenseInt

	mov eax,[newdisk]
	or eax,eax
	je @fdcrec_exit
	call FAT12_LoadFAT
	call FAT12_LoadRootDir
	mov dword[newdisk],0
@fdcrec_exit:
	mov eax,1
	ret

; ebx contains the track to seek to
FDC_Seek:
;	ENTER 4,0		; Reserve 4 bytes on stack
	push ebp
	mov ebp,esp
	add esp,0FFFFFFFCh

	call FDC_MotorOn
	xor eax,eax
	mov [FDC_Done],eax

	mov [ebp-04h],ebx	; save track no.
	mov ebx,CMD_SEEK	; send the seek command
	call FDC_SendCmd	
	xor ebx,ebx		; send zero byte
	call FDC_SendCmd
	mov ebx,[ebp-04h]	; send track no.
	call FDC_SendCmd

	call FDC_WaitDone	; wait for IRQ to finish
	call PrintHex
	call FDC_SenseInt ; and do Sense int

	mov eax,[FDC_Track]
	cmp eax,[ebp-04h]	; check for SE flag
	jne @fdcs_false

	mov eax,1
	jmp @fdcs_exit
@fdcs_false:
	xor eax,eax
@fdcs_exit:
	leave
	ret

; ebx : LBA
; ecx : Buffer
FDC_ReadSector:
	push ebp
	mov ebp,esp
	add esp,0FFFFFFF0h	; Reserve 16 bytes

	mov [ebp-10h],ecx
	call FDC_LBA_to_CHS
	mov [ebp-04h],eax	; save track
	mov [ebp-08h],edx	; save head
	mov [ebp-0Ch],ecx	; save sector

;	call FDC_MotorOn
	mov eax,1
	call FDC_Seek
	call FDC_Recalibrate

	mov ebx,[ebp-04h]
	call FDC_Seek
	or eax,eax
	je @fdcr_bad

	mov edx,FDC_DIR
	xor eax,eax
	in al,dx
	and eax,80h	; check for new disk
	jz @fdcr_1	; if not continue

	mov eax,1
	mov [newdisk],eax
	call FDC_Seek

	call FDC_Recalibrate
	call FDC_MotorOff
@fdcr_bad:
	jmp @fdcr_false

@fdcr_1:
	mov ecx,5	; Try 5 times

@fdcr_readloop:
	push ecx
	xor ebx,ebx
	mov [FDC_Done],ebx
	call DMA_Setup

	mov ebx,CMD_READ
	call FDC_SendCmd

	mov ebx,[ebp-08h]
	shl ebx,2
	call FDC_SendCmd

	mov ebx,[ebp-04h]
	call FDC_SendCmd

	mov ebx,[ebp-08h]
	call FDC_SendCmd

	mov ebx,[ebp-0Ch]
	call FDC_SendCmd

	mov ebx,2		; use 512 byte sector
	call FDC_SendCmd

	mov ebx,DISK_SPT	; send sectors per track
	call FDC_SendCmd

	mov ebx,DISK_GAP3RW
	call FDC_SendCmd

	mov ebx,0FFh
	call FDC_SendCmd

	call FDC_WaitDone

	xor ecx,ecx
	mov edi,status
@fdcr_getstatus:
	push ecx
	cmp ecx,7
	je @fdcr_2
	xor eax,eax
	mov edx,FDC_MSR
	in al,dx
	and eax,0D0h
	cmp eax,0D0h
	jne @fdcr_2

	call FDC_GetByte
	pop ecx
	mov [edi+ecx],al
	inc ecx
	jmp @fdcr_getstatus

@fdcr_2:
	pop ecx	; pop off and extra ecx

	movzx eax,byte[status]
	and eax,0C0h
	or eax,eax
	je @fdcr_ok

	call FDC_Recalibrate
	pop ecx	; get the no. of tries

	dec ecx	; decrement
	jnz @fdcr_gotoloop

	jmp @fdcr_exit
@fdcr_gotoloop:
	jmp @fdcr_readloop
@fdcr_ok:
	mov esi,FDC_DMA_Buffer
	mov edi,[ebp-10h]
	mov ecx,512/4
	rep movsd
@fdcr_true:
	call FDC_MotorOff
	mov eax,1
	jmp @fdcr_exit

@fdcr_false:
	call FDC_MotorOff
	xor eax,eax
@fdcr_exit:
	leave
	ret

; ebx contains the LBA
; Returns
; Track in eax
; Head in edx
; Sector in ecx
FDC_LBA_to_CHS:
	mov eax,ebx		; put LBA in eax for division
	xor edx,edx		; clear edx for division
	mov ecx,DISK_SPT	; divide LBA
	div ecx		; by Sectors per Track
	inc edx		; increment remainder in edx
	mov ecx,edx		; Save Sector = (LBA % SPT)+1

	xor edx,edx		; clear edx for division
	mov ebx,DISK_HEADS; divide (LBA/SPT) by Number of heads
	div ebx		; Then eax : Track no. and edx : Head no.

	ret

; Waits for the IRQ to be completed
FDC_WaitDone:
	mov eax,[FDC_Done]
	or eax,eax
	jz FDC_WaitDone
	ret

; Waits for the IRQ to complete and also sends the Sense Int command
FDC_SenseInt:
	mov ebx,CMD_SENSEINT
	call FDC_SendCmd

	call FDC_GetByte
	mov [ST0_Reg],eax
	call FDC_GetByte
	mov [FDC_Track],eax
	ret

; pass the value to write in ebx
; Routine from Intel manual translated to ASM
FDC_SendCmd:
	push ebp
	mov ebp,esp

	mov ecx,128
@fdwb_1:
	mov edx,FDC_MSR
	in al,dx
	and eax,0C0h
	cmp eax,80h
	jne @fdwb_2

	mov eax,ebx
	mov edx,FDC_DATA
	out dx,al
	jmp @fdwb_exit

@fdwb_2:
	dec ecx
	in al,80h
	jnz @fdwb_1

	mov eax,-1
@fdwb_exit:
	pop ebx
	leave
	ret

FDC_GetByte:
	push ebp
	mov ebp,esp

	xor eax,eax
	mov ecx,128
@fdrb_1:
	mov edx,FDC_MSR
	in al,dx
	and eax,0D0h
	cmp eax,0D0h
	jne @fdrb_2

	xor eax,eax
	mov edx,FDC_DATA
	in al,dx

	jmp @fdrb_exit
@fdrb_2:
	dec ecx
	in al,80h
	jnz @fdrb_1

	mov eax,-1
@fdrb_exit:
	leave
	ret

FDC_Done		dd	0
FDC_Motor_Flag	dd	0
newdisk		dd	0

FDC_Motor_Timer	dd	0
timeout		dd	0

status	times	7 db 0

ST0_Reg	dd	0
FDC_Track	dd	0FFh

FloppyMsg	db	0Eh,"Initializing floppy disk driver.....",0Dh,0
NEC765Msg	db	0Eh,"NEC765A or 8272A Floppy Disk Controller Found",0Dh,0
FDCMsg	db	0Eh,"Enhanced Floppy Disk Controller(82077,82077AA or better) found",0Dh,0
FloppyOk	db	0Eh,"Floppy disk driver OK",0Dh,0Dh,0