[BITS 32]
%define	KEY_LSHIFT	02Ah
%define	KEY_RSHIFT	036h
%define	KEY_ESCAPE	1Bh
%define	KEY_BACKSPACE 08h
%define	KEY_ENTER	0Dh

%define	KEY_UP	80h
%define	KEY_DOWN	81h
%define	KEY_LEFT	82h
%define	KEY_RIGHT	83h

%define	KEY_INS	90h
%define	KEY_DEL	91h
%define	KEY_HOME	92h
%define	KEY_END	93h
%define	KEY_PGUP	94h
%define	KEY_PGDN	95h
%define	KEY_BREAK	96h

%define	KEY_LCTRL	100h
%define	KEY_RCTRL	101h

;Init_Keyboard:
;	mov al,20h
;	out 64h,al	; Tell keyboard we want to read cmd byte
;	call Empty8042
;	in al,60h	; Read in command byte
;	or al,01h	; set bit 0 (keyboard interrupts enabled)
;	;or al,03h	; set bit 0 & 1(keyboard & mouse interrupts enabled)
;	;out 64h,al
;	ret

Keyboard:
	cli
	pushad
	pushf
	xor eax,eax
loop1:
	in al,64h
	and al,01h
	jz loop1
	in al,60h

	mov ebx,eax

	cmp al,0E0h		; if special key
	je @kb_e		; set previous scancode
	cmp al,0E1h		; to it
	je @kb_e		;

	or al,al		;
	je @kb_f		;
	cmp al,0FFh		;
	je @kb_f		;
	
	cmp al,KEY_LSHIFT	; check for left shift
	je @kb_ls		;
	cmp al,KEY_RSHIFT	; check for right shift
	je @kb_rs		;

	cmp al,0AAh		; check for left shift
	je @kb_lS		; release
	cmp al,0B6h		; check for right shift 
	je @kb_rS		; release

	jmp @kb_1

@kb_f:xor al,al
@kb_e:
	mov byte[prev_scode],al
	jmp @kb_0
@kb_s:mov byte[shift],1
	jmp @kb_0
@kb_S:mov byte[shift],0
	jmp @kb_0

@kb_ls:mov byte[lshift],1
	jmp @kb_0
@kb_rs:mov byte[rshift],1
	jmp @kb_0

@kb_lS:mov byte[lshift],0
	jmp @kb_0
@kb_rS:mov byte[rshift],0
	jmp @kb_0

@kb_1:cmp al,02h		; check if keys b/w
	jge $+7		; 02h and 0Dh
	jmp @kb_0		; which are the row of keys above
	cmp al,0Eh		; the alphabets upto backspace
	jle @kb		;

	cmp al,10h		; check if keys b/w
	jge $+7		; 10h and 1Bh
	jmp @kb_0		; which are q w e r t y u i o p [ ]
	cmp al,1Ch		; 
	jle @kb		;

	cmp al,1Eh		; check if keys b/w
	jge $+7		; 1Eh and 28h
	jmp @kb_0		; which are :
	cmp al,28h		; a s d f g h j k l ; '
	jle @kb		;

	cmp al,2Bh		; check if keys b/w
	jge $+7		; 2Ch and 34h
	jmp @kb_0		; which are :
	cmp al,35h		; \ z x c v b n m , . /
	jle @kb		;

	cmp al,39h
	je @kb

	cmp byte[prev_scode],0
	jne @kb_E

	jmp @kb_0
	
@kb:	call getKeyChar
	mov edx,keyBuffer
	add edx,[keyPos]
	mov byte[edx],al
	inc dword[keyBufSize]
	inc dword[keyPos]
	cmp dword[keyBufSize],99
	je short $+7
	jmp @kb1
	mov dword[keyBufSize],0
	mov dword[keyPos],0
@kb1:	jmp @kb_0

@kb_E:mov byte[prev_scode],0
	mov edx,Keymap_E0
	mov al,byte[edx+eax]
	call useSpecKeys
	
@kb_0:call MoveCursor
	mov al,20h
	out 20h,al	; Send EOI (End of Interrupt)

	popf
	popad
	iret

getKeyChar:
	pushf
	push edx

	cmp byte[lshift],1
	je @gk_shift
	cmp byte[rshift],1
	je @gk_shift
	mov edx,Keymap_NoShift
	add dl,al
	jmp @gk_exit
@gk_shift:
	mov edx,Keymap_Shift
	add dl,al
@gk_exit:
	mov al,byte[edx]
;	mov byte[ds:0B8080h],al
;	mov byte[ds:0B8081h],0Dh
	pop edx
	popf
	ret

useSpecKeys:
	cmp al,0Dh
	je short $+7
	jmp @use_exit
	call PutCR
@use_exit:
	ret

Empty8042:
	in al,64h
	test al,1
	jz empty1
	in al,060h
	jmp Empty8042
  empty1:
	test al,2
	jnz Empty8042
	retn 4

Delay32:		; borrowed from SolarOs
	jmp delay_01
delay_01:
	jmp delay_02
delay_02:
	ret

Wait_Read_8042:
	call Delay32

	in al,064h
	and al,01b
	jnz have_data
	jmp Wait_Read_8042
have_data:
	ret

Wait_Write_8042:
	call Delay32

	in al,64h
	and al,10b
	jnz Wait_Write_8042
	ret

kbd0:
	jmp short $+2
	in al,60h
kbd:
	jmp short $+2
	in al,64h
	test al,1
	jnz kbd0
	test al,2
	jnz kbd
	ret

keyBuffer	times 100 db 0
keyPos	dd 0
keyBufSize	dd 0

prev_scode	db 0
shift		db 0
lshift	db 0
rshift	db 0

Keymap_NoShift	db 0,KEY_ESCAPE,"1234567890-=",KEY_BACKSPACE
			db "	qwertyuiop[]",KEY_ENTER,0
			db "asdfghjkl;'`",0,'\'
			db "zxcvbnm,./",0,0,0,' ',0
			db 0,0,0,0,0,0,0,0,0,0
			db 0,0,0,0,0,'-',0,'5',0,'+',0,0
			db 0,0,0,0,0,0,0,0,0

Keymap_Shift	db 0,KEY_ESCAPE, '!@#$%^&*()_+' ,KEY_BACKSPACE
			db '	QWERTYUIOP{}',KEY_ENTER,0
			db 'ASDFGHJKL:"~',0,'|'
			db 'ZXCVBNM<>?',0,0,0,' ',0
			db 0,0,0,0,0,0,0,0,0,0
			db 0,0,0,0,0,'-',0,'5',0,'+',0,0
			db 0,0,0,0,0,0,0,0,0

Keymap_E0	db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, KEY_ENTER, KEY_RCTRL, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, KEY_BREAK, KEY_HOME,
		db KEY_UP, KEY_PGUP, 0, KEY_LEFT, 0, KEY_RIGHT, 0, KEY_END,
		db KEY_DOWN, KEY_PGDN, KEY_INS, KEY_DEL, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0,
		db 0, 0, 0, 0, 0, 0, 0, 0
